/* -*- Mode: C++ -*- */

/*
 *Copyright:

    Copyright (C) 1996-2000 Electrotechnical Laboratory. 
    	Itsuki Noda, Yasuo Kuniyoshi and Hitoshi Matsubara.
    Copyright (C) 2000 RoboCup Soccer Server Maintainance Group.
    	Patrick Riley, Tom Howard, Daniel Polani, Itsuki Noda

    This file is a part of SoccerServer.

    SoccerServer is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

#ifndef _RCC_ACTION_H_
#define _RCC_ACTION_H_

#include <stream.h>
#include <vector>

#include "rccregion.h"
#include "rccgarbage.h"

class RCC_Action : public RCC_GCable
{
public:
  RCC_Action() {}
  virtual ~RCC_Action() {}

  virtual ostream& print ( ostream& out ) const = 0;
  virtual ostream& printPretty ( ostream& out, const string& line_header ) const = 0;
  friend ostream& operator << ( ostream & os, RCC_Action& a )
    { return a.print(os); } 
  
  virtual RCC_Action* copy () const = 0;
};

class RCC_ActRegion : public RCC_Action
{
public:
  RCC_ActRegion ( const RCC_Region* r ) : reg_p ( r->copy () ) {}
  RCC_ActRegion ( const RCC_ActRegion& a ) : reg_p ( a.reg_p->copy () ) {}
  
  ~RCC_ActRegion () { delete reg_p; }
 
  /* class specific stuff */
  RCC_Region* region () const { return reg_p; }
  void region ( const RCC_Region* r ) { delete reg_p; reg_p = r->copy (); }

protected:
  RCC_Region* reg_p;
};


class RCC_ActPosition : public RCC_ActRegion
{
public:
  RCC_ActPosition ( const RCC_Region* r ) : RCC_ActRegion ( r )  {}
  RCC_ActPosition ( const RCC_ActPosition& a ) : RCC_ActRegion ( a )  {}
  ~RCC_ActPosition() {}

  ostream& print ( ostream& out ) const
    { out << "(pos " << *reg_p << ")"; return out; }
  ostream& printPretty ( ostream& out, const string& line_header ) const
    { out << line_header << "position self at: " << endl;
    reg_p->printPretty ( out, line_header + " " ); return out; }

  RCC_Action* copy () const { return ( RCC_Action* ) new RCC_ActPosition ( *this ); }
};



class RCC_ActHome : public RCC_ActRegion
{
public:
  RCC_ActHome ( const RCC_Region* r ) : RCC_ActRegion ( r ) {}
  RCC_ActHome ( const RCC_ActHome& a ) : RCC_ActRegion ( a ) {}
  ~RCC_ActHome() {}

  ostream& print ( ostream& out ) const
    { out << "(home " << *reg_p << ")"; return out; }
  ostream& printPretty ( ostream& out, const string& line_header ) const
    { out << line_header << "home position at: " << endl;
    reg_p->printPretty(out, line_header + " "); return out; }

  RCC_Action* copy () const { return ( RCC_Action* ) new RCC_ActHome ( *this ); }
};


class RCC_ActBallToRegion : public RCC_ActRegion
{
public:
  RCC_ActBallToRegion ( const RCC_Region* r, const RCC_BallMoveSet& bmset ) : RCC_ActRegion ( r ), bm_set ( bmset ) {}
  RCC_ActBallToRegion ( const RCC_ActBallToRegion& a ) : RCC_ActRegion ( a ), bm_set ( a.bm_set ) {}

  ~RCC_ActBallToRegion() {}

  ostream& print ( ostream& out ) const { out << "(bto " << *reg_p << " " << bm_set << ")"; return out; }
  ostream& printPretty ( ostream& out, const string& line_header ) const
    { out << line_header << "ball to: use " << bm_set << " to go to:" << endl;
    reg_p->printPretty ( out, line_header + " " ); return out; }

  /* class specific stuff */
  const RCC_BallMoveSet& BMSet() const { return bm_set; }
  void BMSet ( const RCC_BallMoveSet& bmset ) { bm_set = bmset; }
  void addBMT ( const RCC_BallMoveSet::RCC_BallMove_t& t ) { bm_set.add ( t ); }
  void removeBMT ( const RCC_BallMoveSet::RCC_BallMove_t& t ) { bm_set.remove ( t ); }
  void clearBMT() { bm_set.clear (); }

  RCC_Action* copy () const { return ( RCC_Action* ) new RCC_ActBallToRegion ( *this ); }

protected:
  RCC_BallMoveSet bm_set;
} ;


class RCC_ActBallToPlayer : public RCC_Action
{
public:
  RCC_ActBallToPlayer ( const RCC_UnumSet& p ) : players_ ( p ) {}
  ~RCC_ActBallToPlayer() {}

  ostream& print ( ostream& out ) const { out << "(bto " << players_ << ")"; return out; }
  ostream& printPretty ( ostream& out, const string& line_header ) const
    { out << line_header << "ball to players: " << players_ << endl; return out; }

  /* class specific stuff */
  const RCC_UnumSet& players() const { return players_; }
  void players ( const RCC_UnumSet& players ) { players_ = players; }
  void add ( const int& i ) { players_.add ( i ); }
  void remove ( const int& i ) { players_.remove ( i ); }
  void clear () { players_.clear(); }

  RCC_Action* copy () const { return ( RCC_Action* ) new RCC_ActBallToPlayer ( *this ); }

protected:
  RCC_UnumSet players_;
} ;


class RCC_ActMark : public RCC_Action
{
public:
  RCC_ActMark ( const unsigned int& u ) : unum ( u ) 
  {
    if ( unum == 0 || unum > MAX_PLAYER )
      throw ( "Uniform number out of bounds" ); 
  }

  ~RCC_ActMark() {}

  ostream& print ( ostream& out ) const
    { out << "(mark " << unum << ")"; return out; }
  ostream& printPretty ( ostream& out, const string& line_header ) const
    { out << line_header << "mark opponent " << unum << endl; return out; }

  /* class specific stuff */
  unsigned int uNum () const { return unum; }
  void uNum ( const unsigned int& u ) 
  { unum = u; if ( unum == 0 || unum > MAX_PLAYER ) throw ( "Uniform number out of bounds" ); } 

  RCC_Action* copy () const { return ( RCC_Action* ) new RCC_ActMark ( *this ); }

protected:
  int unum;
};


class RCC_ActMarkLinePlayer : public RCC_ActMark
{
public:
  RCC_ActMarkLinePlayer( const unsigned int& u ) : RCC_ActMark( u ) {}
  ~RCC_ActMarkLinePlayer() {}

  ostream& print ( ostream& out ) const
    { out << "(markl " << unum << ")"; return out; }
  ostream& printPretty ( ostream& out, const string& line_header ) const
    { out << line_header << "mark line to opponent " << unum << endl; return out; }

  RCC_Action* copy () const { return ( RCC_Action* ) new RCC_ActMarkLinePlayer ( *this ); }

};


class RCC_ActMarkLineRegion : public RCC_ActRegion
{
public:
  RCC_ActMarkLineRegion ( const RCC_Region* r ) : RCC_ActRegion ( r ) {}
  RCC_ActMarkLineRegion ( const RCC_ActMarkLineRegion& a ) : RCC_ActRegion ( a ) {}
  ~RCC_ActMarkLineRegion() {}

  ostream& print ( ostream& out ) const
    { out << "(markl " << *reg_p << ")"; return out; }
  ostream& printPretty ( ostream& out, const string& line_header ) const
    { out << line_header << "mark line to region:" << endl;
    reg_p->printPretty ( out, line_header + " " ); return out; }

  RCC_Action* copy () const { return ( RCC_Action* ) new RCC_ActMarkLineRegion ( *this ); }

} ;

class RCC_ActOffsidesLine : public RCC_ActRegion
{
public:
  RCC_ActOffsidesLine ( const RCC_Region* r ) : RCC_ActRegion ( r ) {}
  RCC_ActOffsidesLine ( const RCC_ActOffsidesLine& a ) : RCC_ActRegion ( a ) {}
  ~RCC_ActOffsidesLine() {}

  ostream& print ( ostream& out ) const
    { out << "(oline " << *reg_p << ")"; return out; }
  ostream& printPretty ( ostream& out, const string& line_header ) const
    { out << line_header << "offsides line at: " << endl;
    reg_p->printPretty ( out, line_header + " "); return out; }

  RCC_Action* copy () const { return ( RCC_Action* ) new RCC_ActOffsidesLine ( *this ); }

} ;

class RCC_ActHetType : public RCC_Action
{
public:
  RCC_ActHetType ( const unsigned int& pt ) : player_type ( pt ) {}
  ~RCC_ActHetType() {}

  ostream& print ( ostream& out ) const
    { out << "(htype " << player_type << ")"; return out; }
  ostream& printPretty ( ostream& out, const string& line_header ) const
    { out << line_header << "heterogenous type: " << player_type << endl; return out; }

  /* class specific stuff */
  unsigned int playerType () const { return player_type; }
  void playerType ( const unsigned int& pt ) { player_type = pt; }

  RCC_Action* copy () const { return ( RCC_Action* ) new RCC_ActHetType ( *this ); }

protected:
  int player_type;
};

class RCC_ActNamed : public RCC_Action
{
public:
  RCC_ActNamed ( const string& n ) : name_ ( n ) {}
  ~RCC_ActNamed() {}

  ostream& print ( ostream& out ) const { out << "\"" << name_ << "\""; return out; }
  ostream& printPretty ( ostream& out, const string& line_header ) const
    { out << line_header << "action named \"" << name_ << "\"" << endl; return out; }

  /* class specific */
  string name () const { return name_; }
  void  name ( const string& n ) { name_ = n; }

  RCC_Action* copy () const { return ( RCC_Action* ) new RCC_ActNamed ( *this ); }

protected:
  string name_;
};


#endif
