/* -*- Mode: C++ -*- */
/*
 *Copyright:

    Copyright (C) 2001 RoboCup Soccer Server Maintainance Group.
    	Patrick Riley, Tom Howard, Itsuki Noda,	Mikhail Prokopenko, Jan Wendler 

    This file is part of SoccerServer.

    SoccerServer is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

#ifndef _RCC_COND_H_
#define _RCC_COND_H_

#include <stream.h>
#include <vector>
#include <string>

#include "rccregion.h"
#include "rccplaymode.h"
#include "rccgarbage.h"
#include "rccside.h"

class RCC_Condition : public RCC_GCable
{
public:
  typedef vector < RCC_Condition* > container;

  RCC_Condition() {}
  virtual ~RCC_Condition() {}

  virtual ostream& print ( ostream& out ) const = 0;
  virtual ostream& printPretty ( ostream& out, const string& line_header ) const = 0;

  friend ostream& operator<< ( ostream & os, RCC_Condition& c )
    { return c.print(os); }

  virtual bool eval () const = 0;

  operator bool () const { return eval (); } 

  virtual RCC_Condition* copy () const = 0; 
};

class RCC_CondRegion 
{
public:
  RCC_CondRegion ( const RCC_Region* r ) : reg_p ( r->copy () ) {}
  RCC_CondRegion ( const RCC_CondRegion& c ) : reg_p ( c.reg_p->copy () ) {}

  virtual ~RCC_CondRegion () { delete reg_p; }
 
  /* class specific stuff */
  const RCC_Region* region() const { return reg_p; }
  void region ( const RCC_Region* r ) { delete reg_p; reg_p = r->copy (); }

protected:
  RCC_Region* reg_p;
};

class RCC_CondPlayer 
{
public:
  RCC_CondPlayer ( const RCC_Side& s, const RCC_UnumSet& p )
    : side_ ( s ), players ( p ) {}

  virtual ~RCC_CondPlayer () {}

  /* class specific stuff */
  bool ourSide() const { return side_.ourSide (); }
  bool oppSide() const { return side_.oppSide (); }
  const RCC_UnumSet& playerSet () const { return players; }

  RCC_Side side ( const RCC_Side& s ) { side_ = s; }
  void playerSet ( const RCC_UnumSet& p ) { players = p; }
  void addPlayer ( const int& i ) { players.add ( i ); }
  void removePlayer ( const int& i ) { players.remove ( i ); }
  void clearPlayers () { players.clear(); }

protected:
  RCC_Side side_;
  RCC_UnumSet players;
};

class RCC_CondTrue : public RCC_Condition
{
public:
  RCC_CondTrue() {}
  ~RCC_CondTrue() {}

  ostream& print ( ostream& out ) const { out << "(true)"; return out; }
  ostream& printPretty ( ostream& out, const string& line_header ) const
    { out << line_header << "true" << endl; return out; }

  bool eval () const { return true; }
  
  RCC_Condition* copy () const { return ( RCC_Condition* ) new RCC_CondTrue ( *this ); }

};


class RCC_CondFalse : public RCC_Condition
{
public:
  RCC_CondFalse() {}
  ~RCC_CondFalse() {}

  ostream& print ( ostream& out ) const { out << "(false)"; return out; }
  ostream& printPretty ( ostream& out, const string& line_header ) const
    { out << line_header << "false" << endl; return out; }

  bool eval () const { return false; }

  RCC_Condition* copy () const { return ( RCC_Condition* ) new RCC_CondFalse ( *this ); }

};


class RCC_CondPlayerPosition : public RCC_Condition, public RCC_CondPlayer, public RCC_CondRegion
{
public:
  RCC_CondPlayerPosition ( const RCC_Side& s, const RCC_UnumSet& p, const int& min_players, const int& max_players, const RCC_Region& r )
    : RCC_CondPlayer ( s, p ), RCC_CondRegion ( &r ), min_ ( min_players ), max_ ( max_players ) {}

  ~RCC_CondPlayerPosition() {}

  ostream& print ( ostream& out ) const;
  ostream& printPretty ( ostream& out, const string& line_header ) const;

  bool eval () const { throw ( "eval funtion undefined for RCC_CondPlayerPosition"); } //HOOK: need to add real evaluation here

  RCC_Condition* copy () const { return ( RCC_Condition* ) new RCC_CondPlayerPosition ( *this ); }

  int min () const { return min_; }
  int max () const { return max_; }
  int min ( const int& m ) { return min_ = m; }
  int max ( const int& m ) { return max_ = m; }


protected:
  int min_;
  int max_;
};

class RCC_CondBallPosition : public RCC_Condition, public RCC_CondRegion
{
public:
  RCC_CondBallPosition ( const RCC_Region* r ) : RCC_CondRegion ( r ) {}
  ~RCC_CondBallPosition() {}

  ostream& print ( ostream& out ) const { out << "(bpos " << *reg_p << ")"; return out; }
  ostream& printPretty ( ostream& out, const string& line_header ) const
    { out << line_header << "is ball position at: " << endl;
    reg_p->printPretty(out, line_header + " "); return out; }

  bool eval () const { throw ( "eval funtion undefined for RCC_CondBallPosition"); } //HOOK: need to add real evaluation here

  RCC_Condition* copy () const { return ( RCC_Condition* ) new RCC_CondBallPosition ( *this ); }

};

class RCC_CondBallOwner : public RCC_Condition, public RCC_CondPlayer
{
public:
  RCC_CondBallOwner ( const RCC_Side& s, const RCC_UnumSet& p ) : RCC_CondPlayer ( s, p ) {}
  ~RCC_CondBallOwner() {}

  ostream& print ( ostream& out ) const;
  ostream& printPretty ( ostream& out, const string& line_header ) const;

  bool eval () const { throw ( "eval funtion undefined for RCC_CondBallOwner" ); } //HOOK: need to add real evaluation here

  RCC_Condition* copy () const { return ( RCC_Condition* ) new RCC_CondBallOwner ( *this ); }

};

class RCC_CondPlayMode : public RCC_Condition
{
public:
  RCC_CondPlayMode ( const RCC_PlayMode& pm ) : play_mode ( pm ) {}
  ~RCC_CondPlayMode() {}

  ostream& print ( ostream& out ) const { out << "(playm " << play_mode.coachMode () << ")"; return out; }
  ostream& printPretty ( ostream& out, const string& line_header ) const
    { out << line_header << "is play mode '"
	  << play_mode << "'" << endl; 
    return out; }

  bool eval () const { throw ( "eval funtion undefined for RCC_CondPlayMode" ); } //HOOK: need to add real evaluation here
  
  /* class specific stuff */
  RCC_PlayMode playMode() const { return play_mode; }
  void playMode ( const RCC_PlayMode& pm ) { play_mode = pm; }

  RCC_Condition* copy () const { return ( RCC_Condition* ) new RCC_CondPlayMode ( *this ); }

protected:
  RCC_PlayMode play_mode;
};

class RCC_CondList : public RCC_GCable
{
public:
  RCC_CondList () {}
  ~RCC_CondList() { clear (); }

  void add ( const RCC_Condition* c ) { conds_.push_back ( c->copy () ); }
  const RCC_Condition::container& conds() const { return conds_; }
  void conds ( RCC_Condition::container& c ) 
  { 
    clear (); 
    for ( RCC_Condition::container::iterator i = c.begin (); i != c.end (); i++ )
      add ( *i ); 
  }

  void clear () 
  {
    for ( RCC_Condition::container::iterator i = conds_.begin (); i != conds_.end (); i++ )
      delete *i;
    conds_.clear ();
  }

  ostream& print ( ostream& out ) const { return out << "RCC_CondList"; }

protected:
  RCC_Condition::container conds_;
};

class RCC_CondAnd : public RCC_Condition, public RCC_CondList
{
public:
  RCC_CondAnd () {}
  RCC_CondAnd ( const RCC_CondList& l ) : RCC_CondList ( l ) { }
  ~RCC_CondAnd () {}

  ostream& print ( ostream& out ) const;
  ostream& printPretty ( ostream& out, const string& line_header ) const;

  bool eval () const;

  RCC_Condition* copy () const { return ( RCC_Condition* ) new RCC_CondAnd ( *this ); }

};

class RCC_CondOr : public RCC_Condition, public RCC_CondList
{
public:
  RCC_CondOr () {}
  RCC_CondOr ( const RCC_CondList& l ) : RCC_CondList ( l ) { }
  ~RCC_CondOr () {}

  ostream& print ( ostream& out ) const;
  ostream& printPretty ( ostream& out, const string& line_header ) const;

  bool eval () const;

  RCC_Condition* copy () const { return ( RCC_Condition* ) new RCC_CondOr ( *this ); }

};

class RCC_CondNot : public RCC_Condition
{
public:
  RCC_CondNot ( const RCC_Condition* c ) :  cond_p ( c->copy () ) {}
  ~RCC_CondNot() { delete cond_p;}

  ostream& print ( ostream& out ) const { if (cond_p) out << "(not " << *cond_p << ")"; return out; }
  ostream& printPretty ( ostream& out, const string& line_header ) const
    { out << line_header << "not " << endl;
    cond_p->printPretty ( out, line_header + " +" ); return out; }

  bool eval () const { return !(*cond_p); }
  
  /* class specific stuff */
  RCC_Condition* cond () { return cond_p; }
  void cond ( RCC_Condition* c_p ) { delete cond_p; cond_p = c_p->copy (); }
  void forgetRCC_Condition() { cond_p = NULL; }

  RCC_Condition* copy () const { return ( RCC_Condition* ) new RCC_CondNot ( *this ); }

protected:
  RCC_Condition* cond_p;
};


class RCC_CondNamed : public RCC_Condition
{
public:
  RCC_CondNamed ( string n = string("") ) : name_ ( n ) {}
  ~RCC_CondNamed() {}

  ostream& print ( ostream& out ) const { out << "\"" << name_ << "\""; return out; }
  ostream& printPretty ( ostream& out, const string& line_header ) const
    { out << line_header << "condition named \"" << name_ << "\"" << endl; return out; }

  bool eval () const { throw ( "eval function undefined for RCC_CondNamed"); } //HOOK: add real evlauation

  /* class specific */
  string name () { return name_; }
  void  name ( const string& n ) { name_ = n; }

  RCC_Condition* copy () const { return ( RCC_Condition* ) new RCC_CondNamed ( *this ); }

protected:
  string name_;

};

#endif
