// -*-c++-*-
#ifndef _OBJECT_H_
#define _OBJECT_H_

#include <string>
#include <strstream.h>
#include "rccvector2d.h"
#include "rccangle.h"
#include "rccgarbage.h"

//=======================================================================
//	RCC_Object
//=======================================================================

class RCC_Object : public RCC_GCable 
{
public:
  enum location_t { INVALID_LOC, TOP_LOC, BOTTOM_LOC, LEFT_LOC, RIGHT_LOC, CENTER_LOC, PENALTY_LOC, GOAL_LOC, MAX_LOC };
  static const string LOCATION_NAMES[];

  RCC_Object ( const RCC_Vector2d& pos );
  RCC_Object ( const RCC_Angle& dir );
  virtual ~RCC_Object();
   
  RCC_Vector2d relPos() const { return rel_pos; }
  RCC_Vector2d relPos ( const RCC_Vector2d &pos ) { return ( rel_pos = pos ); }
      
  friend ostream& operator<< ( ostream& o, const RCC_Object& obj ) { return obj.print ( o ); }

  virtual ostream& print ( ostream& o ) const
  { return print ( o, RCC_Vector2d::POLAR, RCC_Angle::DEG ); }
  virtual ostream& print ( ostream& o, RCC_Vector2d::mode_t, RCC_Angle::mode_t ) const = 0;
    
protected:
  RCC_Vector2d rel_pos;		// objects relative position to player
};

//=======================================================================
//	RCC_StatObject
//=======================================================================

class RCC_StatObject : public RCC_Object
{
public:
  RCC_StatObject ( const RCC_Vector2d& pos );
  RCC_StatObject ( const RCC_Angle& dir );
  virtual ~RCC_StatObject();
};

//=======================================================================
//	RCC_Line
//=======================================================================


class RCC_Line : public RCC_StatObject
{
public:
	
  RCC_Line ( const RCC_Object::location_t& name, const RCC_Angle &dir );  
  RCC_Line ( const RCC_Object::location_t& name, const RCC_Vector2d &pos );  
  ~RCC_Line();

  RCC_Object::location_t name () const { return l_name; }
  RCC_Object::location_t name ( const RCC_Object::location_t& ln ) { return ( l_name = ln ); }
    
  virtual ostream& print ( ostream& o, RCC_Vector2d::mode_t = RCC_Vector2d::POLAR, RCC_Angle::mode_t = RCC_Angle::DEG ) const;

protected:
  RCC_Object::location_t l_name;
};

//=======================================================================
//	RCC_MobObject
//=======================================================================

class RCC_MobObject : public RCC_Object
{
public:
  static RCC_Vector2d vel(const RCC_Vector2d &pos, const RCC_Vector2d &chg);
  
  RCC_MobObject ( const RCC_Angle& dir );
  RCC_MobObject ( const RCC_Vector2d& pos );
  RCC_MobObject ( const RCC_Vector2d& pos, const RCC_Vector2d& chg ); 
  virtual ~RCC_MobObject();

  RCC_Vector2d relVel() const { return rel_vel; }
  RCC_Vector2d relVel( const RCC_Vector2d v ) { return ( rel_vel = v ); }
    
  RCC_Vector2d distDirChg() const;
    
protected:
  RCC_Vector2d rel_vel;
};

//=======================================================================
//	RCC_Flag
//=======================================================================

// Note:  	I've decided that flag should inherit from RCC_MobObject, rather than RCC_StatObject
//		as it can have DistChg and DirChg from the players own movement
//		Since a flag can have a relative velocity, it should be a RCC_MobObject...
//		...though I might change my mind in the future :)
 
class RCC_Flag : public RCC_MobObject
{
public: 
  RCC_Flag( const int &name, const RCC_Angle& dir ); 
  RCC_Flag( const int &name, const RCC_Vector2d &pos ); 
  RCC_Flag( const int &name, const RCC_Vector2d &pos, const RCC_Vector2d &chg ); 
  ~RCC_Flag();

  int name () const { return f_name; }
  int name ( const int& fn ) { return ( f_name = fn ); }
    
  virtual ostream& print ( ostream& o, RCC_Vector2d::mode_t = RCC_Vector2d::POLAR, RCC_Angle::mode_t = RCC_Angle::DEG ) const;

  bool isGoal () const { return	( name () == RCC_Flag::getNum ( RCC_Object::GOAL_LOC, RCC_Object::LEFT_LOC ) ||
				  name () == RCC_Flag::getNum ( RCC_Object::GOAL_LOC, RCC_Object::RIGHT_LOC ) ); }

  string getString () const;
  static int getNum ();
  static int getNum ( const RCC_Object::location_t& l);
  static int getNum ( const RCC_Object::location_t& l1, const RCC_Object::location_t& l2 );
  static int getNum ( const RCC_Object::location_t& l1, const int& l2 );
  static int getNum ( const RCC_Object::location_t& l1, const RCC_Object::location_t& l2, const int& l3 );
  static int getNum ( const RCC_Object::location_t& l1, const RCC_Object::location_t& l2, const location_t& l3 );
protected:
  int f_name;    
};


//=======================================================================
//	RCC_Ball
//=======================================================================

class RCC_Ball : public RCC_MobObject
{
public:

  RCC_Ball ( const RCC_Angle &dir );
  RCC_Ball ( const RCC_Vector2d &pos );
  RCC_Ball ( const RCC_Vector2d &pos, const RCC_Vector2d &chg );
  ~RCC_Ball();

  virtual ostream& print ( ostream& o, RCC_Vector2d::mode_t = RCC_Vector2d::POLAR, RCC_Angle::mode_t = RCC_Angle::DEG ) const;

};

//=======================================================================
//	RCC_PlayerName
//=======================================================================

class RCC_PlayerName : public RCC_GCable
{
public:
  static const int UNKNOWN_NUM;
  static const string UNKNOWN_TEAM;

  RCC_PlayerName ( const string& t, const int& u, const bool& g ) : team_name ( t ), unum ( u ), goal_keeper ( g ) {}
  RCC_PlayerName ( const string& t, const int& u ) : team_name ( t ), unum ( u ), goal_keeper ( false ) {}
  RCC_PlayerName ( const string& t ) : team_name ( t ), unum ( RCC_PlayerName::UNKNOWN_NUM ), goal_keeper ( false ) {}
  RCC_PlayerName () : team_name ( RCC_PlayerName::UNKNOWN_TEAM ), unum ( RCC_PlayerName::UNKNOWN_NUM ), goal_keeper ( false ) {}

  bool isKnown () const { return team_name != RCC_PlayerName::UNKNOWN_TEAM; }

  string team() const { return team_name; }
  string team( const string& tname ) { return ( team_name = tname ); }

  bool goalie () const { return goal_keeper; }
  bool goalie ( const bool &gk ) { return ( goal_keeper = gk ); }
    
  int uNum() const { return unum; }
  int uNum( const int& num ) { return ( unum = num ); }
    
  friend ostream& operator<< ( ostream& o, const RCC_PlayerName& pn ) { return pn.print ( o ); }
  ostream& print ( ostream& o ) const;

private:
  string team_name;
  int unum;
  bool goal_keeper;

};

//=======================================================================
//	RCC_Player
//=======================================================================

class RCC_Player : public RCC_MobObject
{
public:
    
  RCC_Player ( const RCC_PlayerName& pn, const RCC_Angle& dir ); 
  RCC_Player ( const RCC_PlayerName& pn, const RCC_Vector2d &pos ); 
  RCC_Player ( const RCC_PlayerName& pn, const RCC_Vector2d &pos, const RCC_Vector2d &chg ); 
  RCC_Player ( const RCC_PlayerName& pn, const RCC_Vector2d &pos, const RCC_Vector2d &chg, const RCC_Angle &body, const RCC_Angle &head ); 
  ~RCC_Player();

  RCC_Angle bodyDir() const { return  body_orientation; }
  RCC_Angle bodyDir( const RCC_Angle &dir ) { return ( body_orientation = dir ); }

  RCC_Angle headDir() const { return head_orientation; }
  RCC_Angle headDir( const RCC_Angle &dir ) { return ( head_orientation = dir ); }

  RCC_PlayerName playerName () const { return p_name; }

  virtual ostream& print ( ostream& o, RCC_Vector2d::mode_t = RCC_Vector2d::POLAR, RCC_Angle::mode_t = RCC_Angle::DEG ) const;

protected:
  RCC_Angle body_orientation;
  RCC_Angle head_orientation;

  RCC_PlayerName p_name;
};

#endif
