/* -*- Mode: C++ -*- */
/*
 *Copyright:

    Copyright (C) 2001 RoboCup Soccer Server Maintainance Group.
    	Patrick Riley, Tom Howard, Itsuki Noda,	Mikhail Prokopenko, Jan Wendler 

    This file is part of SoccerServer.

    SoccerServer is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

/* This files defines a class region to specify regions for the coach messages */

#ifndef _RCC_REGION_H_
#define _RCC_REGION_H_

#include <stream.h>
#include <vector>
#include <string>

#include "rccangle.h"
#include "rccvector2d.h"
#include "rccgarbage.h"
#include "param.h"
#include "rccside.h"

class RCC_Region : public RCC_GCable
{
public:
  typedef vector < RCC_Region* > container;

  RCC_Region() {}
  virtual ~RCC_Region() {};

  virtual ostream& print (ostream& out ) const = 0;
  virtual ostream& printPretty (ostream& out, const string& line_header ) const = 0;
  friend ostream& operator << (ostream & os, const RCC_Region& r )
    { return r.print(os); } 

  virtual RCC_Region* copy() const = 0;
} ;

class RCC_RegNull : public RCC_Region
{
public:
  RCC_RegNull() : RCC_Region() {}
  ~RCC_RegNull() {}

  ostream& print(ostream& out) const 
    { out << "(null)"; return out; }
  ostream& printPretty ( ostream& out, const string& line_header ) const
    { out << line_header << "null region" << endl; return out; }
      
  RCC_Region* copy() const { return ( RCC_Region* ) new RCC_RegNull ( *this ); }
} ;

class RCC_RegPoint : public RCC_Region
{
public:
  virtual ~RCC_RegPoint() {}

  friend ostream& operator << (ostream & os, const RCC_RegPoint& r )
    { return r.print(os); } 

  /* class specific */
    virtual ostream& printSimple ( ostream& out ) const = 0;

  RCC_Region* copy() const { return ( RCC_Region* ) copyPoint (); }
  virtual RCC_RegPoint* copyPoint() const = 0; 
} ;

class RCC_RegAbsPoint : public RCC_RegPoint, public RCC_Vector2d
{
public:
  RCC_RegAbsPoint() {}
  RCC_RegAbsPoint( const double& x_value, const double& y_value ) : RCC_Vector2d ( x_value, y_value ) {}
  ~RCC_RegAbsPoint() {}


  friend ostream& operator << (ostream & os, const RCC_RegAbsPoint& r )
    { return ( ( RCC_RegPoint& ) r ).print(os); } 

  ostream& print ( ostream& out ) const
    { return out << "(pt " << x () << " " << y () << ")"; }
  ostream& printPretty ( ostream& out, const string& line_header ) const 
    { return printSimple ( out << line_header ) << endl; }
  ostream& printSimple ( ostream& out ) const
    { out << "Point(" << x () << ", " << y () << ")"; return out; }

  RCC_RegPoint* copyPoint() const { return ( RCC_RegPoint* ) new RCC_RegAbsPoint ( *this ); } 
} ;

class RCC_RegObjPoint : public RCC_RegPoint
{
public:
  virtual ~RCC_RegObjPoint () {}
};

class RCC_RegBallPoint : public RCC_RegObjPoint
{
public:
  RCC_RegBallPoint () {}
  ~RCC_RegBallPoint () {}


  ostream& print ( ostream& out ) const
    { return out << "(pt ball)"; }
  ostream& printPretty ( ostream& out, const string& line_header ) const 
    { return printSimple ( out << line_header ) << endl; }
  ostream& printSimple ( ostream& out ) const
    { return out << "Point-Ball"; }

  RCC_RegPoint* copyPoint() const { return ( RCC_RegPoint* ) new RCC_RegBallPoint ( *this ); } 
};

class RCC_RegPlayerPoint : public RCC_RegObjPoint
{
public:
  RCC_RegPlayerPoint ( const RCC_Side& s, const int& u ) : side_ ( s ), unum ( u ) {}
  ~RCC_RegPlayerPoint () {}

  ostream& print ( ostream& out ) const
  { return out << "(pt " << side_ << " " << unum << ")"; }
  ostream& printPretty ( ostream& out, const string& line_header ) const 
  { return printSimple ( out << line_header ) << endl; }
  ostream& printSimple ( ostream& out ) const
  { return out << "Point-Player(" << side_ << " " << unum << ")"; }

  RCC_RegPoint* copyPoint() const { return ( RCC_RegPoint* ) new RCC_RegPlayerPoint ( *this ); }

  RCC_Side side () const { return side_; }
  int uNum () const { return unum; }
protected:
  RCC_Side	side_;
  int		unum;
};

class RCC_RegRelPoint : public RCC_RegAbsPoint
{
public:
  RCC_RegRelPoint ( const double& x_value, const double& y_value, const RCC_RegPoint& p )
    : RCC_RegAbsPoint ( x_value, y_value ), point_p ( p.copyPoint () ) {}
  RCC_RegRelPoint ( const RCC_RegRelPoint& p )
    : RCC_RegAbsPoint ( p ), point_p ( p.point_p->copyPoint () ) {}

  ~RCC_RegRelPoint () { delete point_p; }

  ostream& print ( ostream& out ) const
  { return out << "(pt " << x () << " " << y () << " " << point () << ")"; }
  ostream& printPretty ( ostream& out, const string& line_header ) const 
  { return printSimple ( out << line_header ) << endl; }
  ostream& printSimple ( ostream& out ) const
  { return point ().printSimple ( out << "Point-Relative(" << x () << " " << y () ) << ")"; }

  RCC_RegPoint* copyPoint() const { return ( RCC_RegPoint* ) new RCC_RegRelPoint ( *this ); }
  
  const RCC_RegPoint& point () const { return *point_p; }
protected:
  RCC_RegPoint* point_p;
};

class RCC_RegQuad : public RCC_Region
{
public:
  RCC_RegQuad ( const RCC_RegPoint& pt0, 
		const RCC_RegPoint& pt1, 
		const RCC_RegPoint& pt2, 
		const RCC_RegPoint& pt3 ) 
  {
    pts[0] = pt0.copyPoint ();  
    pts[1] = pt1.copyPoint ();
    pts[2] = pt2.copyPoint ();
    pts[3] = pt3.copyPoint ();
  }
  
  RCC_RegQuad ( const RCC_RegQuad& r )
 {
   pts[0] = r.pts[0]->copyPoint (); 
   pts[1] = r.pts[1]->copyPoint (); 
   pts[2] = r.pts[2]->copyPoint (); 
   pts[3] = r.pts[3]->copyPoint ();
 }

  ~RCC_RegQuad() 
  {
    delete pts[0];
    delete pts[1];
    delete pts[2];
    delete pts[3];
  }

  ostream& print ( ostream& out ) const;
  ostream& printPretty ( ostream& out, const string& line_header ) const;

  /* class specific */
  const RCC_RegPoint& point ( const unsigned int& i ) const 
  { 
    if ( i < 4 )
      return *( pts [ i ] );
    else
      throw ( "Point idx out of bounds" );
  }

  void point ( const unsigned int& i, const RCC_RegPoint& pt)
  {
    if ( i < 4 )
      pts [ i ] = pt.copyPoint ();
    else
      throw ( "Point idx out of bounds" );
  }

  RCC_Region* copy() const { return ( RCC_Region* ) new RCC_RegQuad ( *this ); }

protected:
  RCC_RegPoint* pts[4];
};


class RCC_RegArc : public RCC_Region
{
public:
  RCC_RegArc( const RCC_RegPoint& c, const double& sr, const double& er, const RCC_Angle& start_a, const RCC_Angle& span_a )
    : center_p ( c.copyPoint () ), start_rad ( sr ), end_rad ( er ), start_ang ( start_a ), span_ang ( span_a ) 
  {
    if ( start_rad > end_rad )
      throw ( "Start radius greater than end radius" );
  } 

  RCC_RegArc( const RCC_RegArc& r )
    : center_p ( r.center_p->copyPoint () ), start_rad ( r.start_rad ), end_rad ( r.end_rad ), 
    start_ang ( r.start_ang ), span_ang ( r.span_ang )
  {}

  ~RCC_RegArc() {}

  ostream& print ( ostream& out ) const;
  ostream& printPretty ( ostream& out, const string& line_header ) const;

  /* class specific */
  const RCC_RegPoint& center () const { return *center_p; }
  double startRad () const { return start_rad; }
  double endRad () const  { return end_rad; }
  RCC_Angle startAng () const { return start_ang; }
  RCC_Angle spanAng () const  { return span_ang; }

  void center ( const RCC_RegPoint& pt) { delete center_p; center_p = pt.copyPoint (); }
  void rad ( const double& sr, const double& er );
  void ang ( const RCC_Angle& start_a, const RCC_Angle& span_a );

  RCC_Region* copy() const { return ( RCC_Region* ) new RCC_RegArc ( *this ); }
  
protected:
  RCC_RegPoint* center_p;
  /* start rad <= end_rad */
  double start_rad, end_rad;
  RCC_Angle start_ang, span_ang;
} ;


class RCC_RegUnion : public RCC_Region

{
public:
  RCC_RegUnion () {}
  RCC_RegUnion ( const RCC_RegUnion& r );
  ~RCC_RegUnion() { clear (); }

  ostream& print ( ostream& out ) const;
  ostream& printPretty ( ostream& out, const string& line_header ) const;

  /* class specific */
  int numRegions() const { return regions_.size(); }
  RCC_Region* region ( const unsigned int& i );
  void add ( const RCC_Region* r );
  void remove ( const unsigned int& i );
  void clear ();

  RCC_Region* copy() const { return ( RCC_Region* ) new RCC_RegUnion ( *this ); }
  
protected:
  RCC_Region::container regions_;
} ;

class RCC_RegNamed : public RCC_Region
{
public:
  RCC_RegNamed ( const string& n ) : name_ ( n ) {}
  ~RCC_RegNamed () {}

  ostream& print(ostream& out) const
    { out << "\"" << name_ << "\""; return out; }
  ostream& printPretty(ostream& out, const string& line_header) const
    { out << line_header << "region named \"" << name_ << "\"" << endl; return out; }

  /* class specific */
  string name () { return name_; }
  void  name ( const string& n ) { name_ = n; }

  RCC_Region* copy() const { return ( RCC_Region* ) new RCC_RegNamed ( *this ); }

protected:
  string name_;
} ;





class RCC_BallMoveSet : public RCC_GCable
{
public:
  enum RCC_BallMove_t { None, Pass, Dribble, Clear, Score, All };
  static const string NAMES[];
  static const unsigned int MAX;

  RCC_BallMoveSet ();
  RCC_BallMoveSet ( const RCC_BallMoveSet::RCC_BallMove_t& init_entry );
  ~RCC_BallMoveSet ();

  void clear() { entries = 0; }
  void add ( const RCC_BallMoveSet::RCC_BallMove_t& t );
  void remove ( const RCC_BallMoveSet::RCC_BallMove_t& t );
  bool isMember ( const RCC_BallMoveSet::RCC_BallMove_t& t );

  ostream& print ( ostream& out ) const;
  friend ostream& operator <<(ostream & os, const RCC_BallMoveSet& s )
    { return s.print(os); }
 
protected:
  unsigned entries;
};

class RCC_UnumSet : public RCC_GCable
{
public:
  RCC_UnumSet ( int ts = MAX_PLAYER );
  ~RCC_UnumSet ();

  static const unsigned int ALL;

  void clear() { entries = 0; }
  void add ( const unsigned int& i );
  void remove ( const unsigned int& i );
  bool isMember ( const unsigned int& i ) const;

  ostream& print ( ostream& out ) const;
  friend ostream& operator <<( ostream & os, const RCC_UnumSet& s )
    { return s.print ( os ); }
 
protected:
  unsigned int team_size;
  unsigned int entries;
};


#endif
